@@ -27,6 +27,21 @@
 -- No.2-2-5
 \dS+ dbms_stats.relation_stats_locked
 
+/*
+ * No.2-4 dbms_stats.anyarray
+ */
+-- No.2-4-1
+SELECT n.nspname, t.typname, t.typlen, t.typbyval, t.typtype,
+       t.typcategory, t.typispreferred, t.typisdefined, t.typdelim,
+       t.typrelid, t.typelem, t.typinput, t.typoutput, t.typreceive,
+       t.typsend, t.typmodin, t.typmodout, t.typanalyze, t.typalign,
+       t.typstorage, t.typnotnull, t.typbasetype, t.typtypmod, t.typndims,
+       t.typcollation, t.typdefaultbin, t.typdefault, t.typacl
+  FROM pg_type t, pg_namespace n
+ WHERE t.typnamespace = n.oid
+   AND n.nspname = 'dbms_stats'
+   AND t.typname = 'anyarray';
+
 /*
  * No.5-1 dbms_stats.merge
  */
@@ -857,10 +872,10 @@ SELECT dbms_stats.unlock_database_stats();
 /*
  * No.6-4 dbms_stats.is_target_relkind
  */
--- No.6-4-10
-SELECT dbms_stats.is_target_relkind('f');
--- No.6-4-11
-SELECT dbms_stats.is_target_relkind('m');
+-- No.6-4- 10-11
+SELECT dbms_stats.is_target_relkind(k::"char")
+ FROM (VALUES ('r'), ('i'), ('f'), ('m'),
+ 			  ('S'), ('t'), ('v'), ('c')) t(k);
 
 /*
  * No.7-1 dbms_stats.backup
@@ -1137,16 +1152,11 @@ VACUUM ANALYZE;
  */
 -- No.9-1-1
 DELETE FROM dbms_stats._relation_stats_locked;
+VACUUM dbms_stats._relation_stats_locked;  -- in order to avoid auto vacuum
 BEGIN;
-SELECT relation::regclass, mode
-  FROM pg_locks
- WHERE mode LIKE '%ExclusiveLock%'
- ORDER BY relation::regclass::text, mode;
+SELECT * FROM internal_locks;
 SELECT dbms_stats.restore(2, 's0.st0', NULL);
-SELECT relation::regclass, mode
-  FROM pg_locks
- WHERE mode LIKE '%ExclusiveLock%'
- ORDER BY relation::regclass::text, mode;
+SELECT * FROM internal_locks;
 COMMIT;
 SELECT relid::regclass FROM dbms_stats.relation_stats_locked
  GROUP BY relid
@@ -1248,10 +1258,10 @@ SELECT starelid, attname, stainherit FROM columns_locked_v c;
 DELETE FROM dbms_stats._relation_stats_locked;
 \set s1_st0_oid `psql contrib_regression -tA -c "SELECT c.oid FROM pg_class c, pg_namespace n WHERE c.relnamespace = n.oid AND n.nspname = 's1' AND c.relname = 'st0';"`
 DROP TABLE s1.st0;
-SET client_min_messages TO FATAL;
-SELECT dbms_stats.restore(2, :s1_st0_oid, NULL);
-RESET client_min_messages;
-\! tail ${PGDATA}/pg_log/postgresql.log > results/ut_no3_1_14.out
+-- SELECT dbms_stats.restore(2, :s1_st0_oid, NULL);
+-- To avoid test unstability caused by relation id alloction, the test
+-- above is omitted.
+
 SELECT count(*) FROM dbms_stats.relation_stats_locked;
 SELECT count(*) FROM dbms_stats.column_stats_locked;
 CREATE TABLE s1.st0(id integer, num integer);
@@ -1485,16 +1495,11 @@ SELECT dbms_stats.restore_stats(0);
 
 -- No.10-7-3
 DELETE FROM dbms_stats._relation_stats_locked;
+VACUUM dbms_stats._relation_stats_locked;  -- in order to avoid auto vacuum
 BEGIN;
-SELECT relation::regclass, mode
-  FROM pg_locks
- WHERE mode LIKE '%ExclusiveLock%'
- ORDER BY relation::regclass::text, mode;
+SELECT * FROM internal_locks;
 SELECT dbms_stats.restore_stats(2);
-SELECT relation::regclass, mode
-  FROM pg_locks
- WHERE mode LIKE '%ExclusiveLock%'
- ORDER BY relation::regclass::text, mode;
+SELECT * FROM internal_locks;
 COMMIT;
 SELECT relid::regclass FROM dbms_stats.relation_stats_locked
  GROUP BY relid
@@ -1648,18 +1653,13 @@ SELECT starelid, attname, stainherit FROM columns_locked_v c;
  */
 -- No.11-2-1
 DELETE FROM dbms_stats._relation_stats_locked;
+VACUUM dbms_stats._relation_stats_locked;  -- in order to avoid auto vacuum
 BEGIN;
-SELECT relation::regclass, mode
-  FROM pg_locks
- WHERE mode LIKE '%ExclusiveLock%'
- ORDER BY relation::regclass::text, mode;
+SELECT * FROM internal_locks;
 SELECT dbms_stats.lock('s0.st0');
 SELECT * FROM relations_locked_v;
 SELECT * FROM columns_locked_v c;
-SELECT relation::regclass, mode
-  FROM pg_locks
- WHERE mode LIKE '%ExclusiveLock%'
- ORDER BY relation::regclass::text, mode;
+SELECT * FROM internal_locks;
 COMMIT;
 
 -- No.11-2-2
@@ -1928,7 +1928,7 @@ SELECT relid::regclass FROM dbms_stats._relation_stats_locked
  ORDER BY relid;
 SELECT starelid::regclass, staattnum FROM dbms_stats._column_stats_locked
  GROUP BY starelid, staattnum
- ORDER BY starelid;
+ ORDER BY starelid, staattnum;
 
 -- No.13-1-11
 DELETE FROM dbms_stats._relation_stats_locked;
@@ -1938,14 +1938,14 @@ SELECT relid::regclass FROM dbms_stats._relation_stats_locked
  ORDER BY relid;
 SELECT starelid::regclass, staattnum FROM dbms_stats._column_stats_locked
  GROUP BY starelid, staattnum
- ORDER BY starelid;
+ ORDER BY starelid, staattnum;
 SELECT dbms_stats.unlock(NULL, 'id');
 SELECT relid::regclass FROM dbms_stats._relation_stats_locked
  GROUP BY relid
  ORDER BY relid;
 SELECT starelid::regclass, staattnum FROM dbms_stats._column_stats_locked
  GROUP BY starelid, staattnum
- ORDER BY starelid;
+ ORDER BY starelid, staattnum;
 
 -- No.13-1-12
 DELETE FROM dbms_stats._relation_stats_locked;
@@ -1955,30 +1955,25 @@ SELECT relid::regclass FROM dbms_stats._relation_stats_locked
  ORDER BY relid;
 SELECT starelid::regclass, staattnum FROM dbms_stats._column_stats_locked
  GROUP BY starelid, staattnum
- ORDER BY starelid;
+ ORDER BY starelid, staattnum;
 SELECT dbms_stats.unlock('s0.st0', NULL);
 SELECT relid::regclass FROM dbms_stats._relation_stats_locked
  GROUP BY relid
  ORDER BY relid;
 SELECT starelid::regclass, staattnum FROM dbms_stats._column_stats_locked
  GROUP BY starelid, staattnum
- ORDER BY starelid;
+ ORDER BY starelid, staattnum;
 
 -- No.13-1-13
 DELETE FROM dbms_stats._relation_stats_locked;
+VACUUM dbms_stats._relation_stats_locked;  -- in order to avoid auto vacuum
 SELECT dbms_stats.lock_database_stats();
 SELECT count(*) FROM dbms_stats._relation_stats_locked;
 SELECT count(*) FROM dbms_stats._column_stats_locked;
 BEGIN;
-SELECT relation::regclass, mode
-  FROM pg_locks
- WHERE mode LIKE '%ExclusiveLock%'
- ORDER BY relation::regclass::text, mode;
+SELECT * FROM internal_locks;
 SELECT dbms_stats.unlock();
-SELECT relation::regclass, mode
-  FROM pg_locks
- WHERE mode LIKE '%ExclusiveLock%'
- ORDER BY relation::regclass::text, mode;
+SELECT * FROM internal_locks;
 COMMIT;
 SELECT count(*) FROM dbms_stats._relation_stats_locked;
 SELECT count(*) FROM dbms_stats._column_stats_locked;
@@ -2020,19 +2015,14 @@ SELECT count(*) FROM dbms_stats._column_stats_locked;
 
 -- No.14-1-4
 DELETE FROM dbms_stats._relation_stats_locked;
+VACUUM dbms_stats._relation_stats_locked;  -- in order to avoid auto vacuum
 SELECT dbms_stats.lock_database_stats();
 SELECT count(*) FROM dbms_stats._relation_stats_locked;
 SELECT count(*) FROM dbms_stats._column_stats_locked;
 BEGIN;
-SELECT relation::regclass, mode
-  FROM pg_locks
- WHERE mode LIKE '%ExclusiveLock%'
- ORDER BY relation::regclass::text, mode;
+SELECT * FROM internal_locks;
 SELECT dbms_stats.unlock_database_stats();
-SELECT relation::regclass, mode
-  FROM pg_locks
- WHERE mode LIKE '%ExclusiveLock%'
- ORDER BY relation::regclass::text, mode;
+SELECT * FROM internal_locks;
 COMMIT;
 SELECT count(*) FROM dbms_stats._relation_stats_locked;
 SELECT count(*) FROM dbms_stats._column_stats_locked;
@@ -2169,6 +2159,7 @@ SELECT starelid::regclass, count(*) FROM dbms_stats._column_stats_locked
 
 -- No.14-2-8
 DELETE FROM dbms_stats._relation_stats_locked;
+VACUUM dbms_stats._relation_stats_locked;  -- in order to avoid auto vacuum
 SELECT dbms_stats.lock_database_stats();
 SELECT relid::regclass FROM dbms_stats._relation_stats_locked
  GROUP BY relid
@@ -2177,15 +2168,9 @@ SELECT starelid::regclass, count(*) FROM dbms_stats._column_stats_locked
  GROUP BY starelid
  ORDER BY starelid;
 BEGIN;
-SELECT relation::regclass, mode
-  FROM pg_locks
- WHERE mode LIKE '%ExclusiveLock%'
- ORDER BY relation::regclass::text, mode;
+SELECT * FROM internal_locks;
 SELECT dbms_stats.unlock_schema_stats('s0');
-SELECT relation::regclass, mode
-  FROM pg_locks
- WHERE mode LIKE '%ExclusiveLock%'
- ORDER BY relation::regclass::text, mode;
+SELECT * FROM internal_locks;
 COMMIT;
 SELECT relid::regclass FROM dbms_stats._relation_stats_locked
  GROUP BY relid
@@ -2326,6 +2311,7 @@ SELECT starelid::regclass, count(*) FROM dbms_stats._column_stats_locked
 
 -- No.14-3-8
 DELETE FROM dbms_stats._relation_stats_locked;
+VACUUM dbms_stats._relation_stats_locked;  -- in order to avoid auto vacuum
 SELECT dbms_stats.lock_database_stats();
 SELECT relid::regclass FROM dbms_stats._relation_stats_locked
  GROUP BY relid
@@ -2334,15 +2320,9 @@ SELECT starelid::regclass, count(*) FROM dbms_stats._column_stats_locked
  GROUP BY starelid
  ORDER BY starelid;
 BEGIN;
-SELECT relation::regclass, mode
-  FROM pg_locks
- WHERE mode LIKE '%ExclusiveLock%'
- ORDER BY relation::regclass::text, mode;
+SELECT * FROM internal_locks;
 SELECT dbms_stats.unlock_table_stats('s0.st0');
-SELECT relation::regclass, mode
-  FROM pg_locks
- WHERE mode LIKE '%ExclusiveLock%'
- ORDER BY relation::regclass::text, mode;
+SELECT * FROM internal_locks;
 COMMIT;
 SELECT relid::regclass FROM dbms_stats._relation_stats_locked
  GROUP BY relid
@@ -2483,6 +2463,7 @@ SELECT starelid::regclass, count(*) FROM dbms_stats._column_stats_locked
 
 -- No.14-4-8
 DELETE FROM dbms_stats._relation_stats_locked;
+VACUUM dbms_stats._relation_stats_locked;  -- in order to avoid auto vacuum
 SELECT dbms_stats.lock_database_stats();
 SELECT relid::regclass FROM dbms_stats._relation_stats_locked
  GROUP BY relid
@@ -2491,15 +2472,9 @@ SELECT starelid::regclass, count(*) FROM dbms_stats._column_stats_locked
  GROUP BY starelid
  ORDER BY starelid;
 BEGIN;
-SELECT relation::regclass, mode
-  FROM pg_locks
- WHERE mode LIKE '%ExclusiveLock%'
- ORDER BY relation::regclass::text, mode;
+SELECT * FROM internal_locks;
 SELECT dbms_stats.unlock_table_stats('s0', 'st0');
-SELECT relation::regclass, mode
-  FROM pg_locks
- WHERE mode LIKE '%ExclusiveLock%'
- ORDER BY relation::regclass::text, mode;
+SELECT * FROM internal_locks;
 COMMIT;
 SELECT relid::regclass FROM dbms_stats._relation_stats_locked
  GROUP BY relid
@@ -2625,21 +2600,16 @@ SELECT relid::regclass FROM dbms_stats._relation_stats_locked
 
 -- No.14-5-9
 DELETE FROM dbms_stats._relation_stats_locked;
+VACUUM dbms_stats._relation_stats_locked;  -- in order to avoid auto vacuum
 SELECT dbms_stats.lock_database_stats();
 SELECT relid::regclass FROM dbms_stats._relation_stats_locked
  GROUP BY relid
  ORDER BY relid;
 SELECT starelid, attname, stainherit FROM columns_locked_v c;
 BEGIN;
-SELECT relation::regclass, mode
-  FROM pg_locks
- WHERE mode LIKE '%ExclusiveLock%'
- ORDER BY relation::regclass::text, mode;
+SELECT * FROM internal_locks;
 SELECT dbms_stats.unlock_column_stats('s0.st0', 'id');
-SELECT relation::regclass, mode
-  FROM pg_locks
- WHERE mode LIKE '%ExclusiveLock%'
- ORDER BY relation::regclass::text, mode;
+SELECT * FROM internal_locks;
 COMMIT;
 SELECT starelid, attname, stainherit FROM columns_locked_v c;
 SELECT relid::regclass FROM dbms_stats._relation_stats_locked
@@ -2748,21 +2718,16 @@ SELECT relid::regclass FROM dbms_stats._relation_stats_locked
 
 -- No.14-6-8
 DELETE FROM dbms_stats._relation_stats_locked;
+VACUUM dbms_stats._relation_stats_locked;  -- in order to avoid auto vacuum
 SELECT dbms_stats.lock_database_stats();
 SELECT relid::regclass FROM dbms_stats._relation_stats_locked
  GROUP BY relid
  ORDER BY relid;
 SELECT starelid, attname, stainherit FROM columns_locked_v c;
 BEGIN;
-SELECT relation::regclass, mode
-  FROM pg_locks
- WHERE mode LIKE '%ExclusiveLock%'
- ORDER BY relation::regclass::text, mode;
+SELECT * FROM internal_locks;
 SELECT dbms_stats.unlock_column_stats('s0', 'st0', 'id');
-SELECT relation::regclass, mode
-  FROM pg_locks
- WHERE mode LIKE '%ExclusiveLock%'
- ORDER BY relation::regclass::text, mode;
+SELECT * FROM internal_locks;
 COMMIT;
 SELECT starelid, attname, stainherit FROM columns_locked_v c;
 SELECT relid::regclass FROM dbms_stats._relation_stats_locked
