@@ -1,5 +1,5 @@
 \pset null '(null)'
-LOAD 'pg_dbms_stats';
+
 /*
  * No.2-3 dbms_stats.backup_history_id_seq
  */
@@ -20,20 +20,6 @@ SELECT id, time, unit FROM dbms_stats.backup_history
 SELECT setval('dbms_stats.backup_history_id_seq', 1, false);
 DELETE FROM dbms_stats.backup_history;
 
-/*
- * No.2-4 dbms_stats.anyarray
- */
--- No.2-4-1
-SELECT n.nspname, t.typname, t.typlen, t.typbyval, t.typtype,
-       t.typcategory, t.typispreferred, t.typisdefined, t.typdelim,
-       t.typrelid, t.typelem, t.typinput, t.typoutput, t.typreceive,
-       t.typsend, t.typmodin, t.typmodout, t.typanalyze, t.typalign,
-       t.typstorage, t.typnotnull, t.typbasetype, t.typtypmod, t.typndims,
-       t.typcollation, t.typdefaultbin, t.typdefault, t.typacl
-  FROM pg_type t, pg_namespace n
- WHERE t.typnamespace = n.oid
-   AND n.nspname = 'dbms_stats'
-   AND t.typname = 'anyarray';
 /*
  * No.3-1 dbms_stats.use_locked_stats
  */
@@ -41,15 +27,15 @@ DELETE FROM dbms_stats._relation_stats_locked;
 EXPLAIN (costs false) SELECT * FROM s0.st2 WHERE id < 1;
 SELECT dbms_stats.lock_table_stats('s0.st2'::regclass);
 UPDATE dbms_stats._relation_stats_locked SET curpages = 10000;
+VACUUM dbms_stats._relation_stats_locked;  -- in order to avoid auto vacuum
 -- No.3-1-1
-SET pg_dbms_stats.use_locked_stats TO on;
+SET pg_dbms_stats.use_locked_stats TO ON;
 EXPLAIN (costs false) SELECT * FROM s0.st2 WHERE id < 1;
 -- No.3-1-2
-SET pg_dbms_stats.use_locked_stats TO off;
+SET pg_dbms_stats.use_locked_stats TO OFF;
 EXPLAIN (costs false) SELECT * FROM s0.st2 WHERE id < 1;
 -- No.3-1-3
 \c - regular_user
-LOAD '$libdir/plugins/pg_dbms_stats';
 SHOW pg_dbms_stats.use_locked_stats;
 SET pg_dbms_stats.use_locked_stats TO OFF;
 SHOW pg_dbms_stats.use_locked_stats;
@@ -57,8 +43,7 @@ EXPLAIN (costs false) SELECT * FROM s0.st2 WHERE id < 1;
 RESET pg_dbms_stats.use_locked_stats;
 EXPLAIN (costs false) SELECT * FROM s0.st2 WHERE id < 1;
 -- clean up
-\c - postgres
-LOAD 'pg_dbms_stats';
+\c - super_user
 DELETE FROM dbms_stats._relation_stats_locked;
 
 /*
@@ -67,8 +52,9 @@ DELETE FROM dbms_stats._relation_stats_locked;
 CREATE TABLE st3(id integer, name char(1000), num_arr char(5)[]);
 INSERT INTO st3 SELECT i, i , ARRAY[i::char, 'a'] FROM generate_series(1,10) g(i);
 ANALYZE st3;
-SELECT stavalues1 FROM pg_statistic
- WHERE starelid = 'public.st3'::regclass;
+SELECT staattnum, stavalues1 FROM pg_statistic
+ WHERE starelid = 'public.st3'::regclass
+ ORDER BY staattnum;
 \copy (SELECT stavalues1::dbms_stats.anyarray FROM dbms_stats.column_stats_effective WHERE starelid = 'st3'::regclass) TO 'results/anyarray_test.cp' binary
 CREATE TABLE st4 (arr dbms_stats.anyarray);
 
@@ -153,7 +139,7 @@ SELECT dbms_stats.unlock_database_stats();
 SELECT dbms_stats.lock_table_stats('st1');
 
 /*
- * No.5-2 dbms_stats.invalidate_column_cache
+ * No.5-2 invalid calls of dbms_stats.invalidate_column_cache
  */
 -- No.5-2-1
 SELECT dbms_stats.invalidate_column_cache();
@@ -213,6 +199,7 @@ UPDATE dbms_stats._column_stats_locked SET stanullfrac = 1
  WHERE starelid = 'st1'::regclass
    AND staattnum = 1
    AND stainherit = false;
+VACUUM dbms_stats._relation_stats_locked;  -- in order to avoid auto vacuum
 EXPLAIN (costs false) SELECT * FROM st1 WHERE val IS NULL;
 
 -- No.5-2-10
@@ -307,6 +294,7 @@ EXPLAIN (costs false) SELECT * FROM st1 WHERE val IS NULL;
 EXPLAIN (costs false) SELECT * FROM st1 WHERE val IS NULL;
 UPDATE dbms_stats._relation_stats_locked SET curpages = 1
  WHERE relid = 'st1'::regclass;
+VACUUM dbms_stats._relation_stats_locked;  -- in order to avoid auto vacuum
 EXPLAIN (costs false) SELECT * FROM st1 WHERE val IS NULL;
 
 -- No.5-3-10
@@ -327,11 +315,13 @@ SELECT dbms_stats.lock_table_stats('st1');
 SELECT relname, curpages FROM dbms_stats.relation_stats_locked
  WHERE relid = 'st1'::regclass;
 SELECT pg_sleep(0.5);
-SELECT pg_stat_reset();
+SELECT reset_stat_and_cache();
+VACUUM dbms_stats._relation_stats_locked;  -- in order to avoid auto vacuum
 UPDATE dbms_stats._relation_stats_locked SET curpages = 1000
  WHERE relid = 'st1_exp'::regclass;
 SELECT pg_sleep(0.5);
 SELECT * FROM lockd_io;
+SELECT reset_stat_and_cache();
 SELECT relname, curpages FROM dbms_stats.relation_stats_locked
  WHERE relid = 'st1'::regclass;
 SELECT pg_sleep(0.5);
@@ -343,10 +333,12 @@ SELECT * FROM lockd_io;
 -- No.5-4-1
 UPDATE dbms_stats._relation_stats_locked SET curpages = 1
  WHERE relid = 'st1'::regclass;
+VACUUM dbms_stats._relation_stats_locked;  -- in order to avoid auto vacuum
 EXPLAIN (costs false) SELECT * FROM st1 WHERE val IS NULL;
 \c
+SET pg_dbms_stats.use_locked_stats to NO;
 EXPLAIN (costs false) SELECT * FROM st1 WHERE val IS NULL;
-LOAD 'pg_dbms_stats';
+SET pg_dbms_stats.use_locked_stats to YES;
 EXPLAIN (costs false) SELECT * FROM st1 WHERE val IS NULL;
 
 SELECT dbms_stats.unlock_database_stats();
@@ -375,10 +367,11 @@ VACUUM ANALYZE;
 SELECT * FROM s0.droptest
  WHERE id = 1;
 SELECT pg_sleep(0.5);
-SELECT pg_stat_reset();
+SELECT reset_stat_and_cache();
 ALTER TABLE s0.droptest RENAME TO test;
 SELECT pg_sleep(0.5);
 SELECT * FROM lockd_io;
+SELECT reset_stat_and_cache();
 SELECT * FROM s0.test
  WHERE id = 1;
 SELECT pg_sleep(0.5);
@@ -389,10 +382,11 @@ ALTER TABLE s0.test RENAME TO droptest;
 SELECT * FROM s0.droptest
  WHERE id = 1;
 SELECT pg_sleep(0.5);
-SELECT pg_stat_reset();
+SELECT reset_stat_and_cache();
 ALTER TABLE s0.droptest RENAME id TO test;
 SELECT pg_sleep(0.5);
 SELECT * FROM lockd_io;
+SELECT reset_stat_and_cache();
 SELECT * FROM s0.droptest
  WHERE test = 1;
 SELECT pg_sleep(0.5);
@@ -404,10 +398,11 @@ INSERT INTO s0.droptest VALUES (4);
 SELECT * FROM s0.droptest
  WHERE id = 1;
 SELECT pg_sleep(0.5);
-SELECT pg_stat_reset();
+SELECT reset_stat_and_cache();
 ANALYZE;
 SELECT pg_sleep(0.5);
 SELECT * FROM lockd_io;
+SELECT reset_stat_and_cache();
 SELECT * FROM s0.droptest
  WHERE id = 1;
 SELECT pg_sleep(1.0);
@@ -419,10 +414,11 @@ INSERT INTO s0.droptest VALUES (4),(5);
 SELECT * FROM s0.droptest
  WHERE id = 4;
 SELECT pg_sleep(0.5);
-SELECT pg_stat_reset();
+SELECT reset_stat_and_cache();
 VACUUM;
 SELECT pg_sleep(0.5);
 SELECT * FROM lockd_io;
+SELECT reset_stat_and_cache();
 SELECT * FROM s0.droptest
  WHERE id = 4;
 SELECT pg_sleep(0.5);
@@ -644,10 +640,16 @@ SELECT starelid::regclass, staattnum FROM dbms_stats.column_stats_backup
  GROUP BY starelid, staattnum
  ORDER BY starelid, staattnum;
 SELECT count(*) FROM dbms_stats.relation_stats_backup;
-SET client_min_messages TO FATAL;
-SELECT dbms_stats.backup(1, 's0.st0'::regclass, NULL);
-RESET client_min_messages;
-\! tail ${PGDATA}/pg_log/postgresql.log > results/ut_no2_1_15.out
+-- SELECT dbms_stats.backup(1, 's0.st0'::regclass, NULL);
+-- To avoid test unstability caused by relation id allocation, unique
+-- constraint which used to be checked above is now checked more
+-- directly in the following step.
+SELECT ic.relname idxname, i.indisprimary
+ FROM pg_index i
+ JOIN pg_class c ON (c.oid = i.indrelid)
+ JOIN pg_namespace n ON (n.oid = c.relnamespace)
+ JOIN pg_class ic ON (ic.oid = i.indexrelid)
+ WHERE n.nspname = 'dbms_stats' AND c.relname = 'relation_stats_backup';
 SELECT count(*) FROM dbms_stats.relation_stats_backup;
 SELECT count(*) FROM dbms_stats.column_stats_backup;
 
@@ -912,13 +914,17 @@ SELECT * FROM dbms_stats.backup_history;
 BEGIN;
 SELECT relation::regclass, mode
   FROM pg_locks
- WHERE mode LIKE '%ExclusiveLock%'
- ORDER BY relation::regclass::text, mode;
+  WHERE relation::regclass::text LIKE 'dbms_stats.\_%\_locked'
+     OR relation::regclass::text LIKE 'dbms_stats.backup_history'
+     OR relation::regclass::text LIKE 'dbms_stats.%\_backup'
+  ORDER BY relation::regclass::text, mode;
 SELECT id, unit, comment FROM dbms_stats.purge_stats(2);
 SELECT relation::regclass, mode
   FROM pg_locks
- WHERE mode LIKE '%ExclusiveLock%'
- ORDER BY relation::regclass::text, mode;
+  WHERE relation::regclass::text LIKE 'dbms_stats.\_%\_locked'
+     OR relation::regclass::text LIKE 'dbms_stats.backup_history'
+     OR relation::regclass::text LIKE 'dbms_stats.%\_backup'
+  ORDER BY relation::regclass::text, mode;
 COMMIT;
 SELECT * FROM dbms_stats.backup_history;
 -- No.15-1-6
@@ -1131,5 +1137,4 @@ SELECT count(*) FROM dbms_stats.stats WHERE false;
 SELECT count(*) FROM dbms_stats._relation_stats_locked WHERE false;
 -- No.19-1-5
 SELECT count(*) FROM dbms_stats._columnns_user WHERE false;
-\c - postgres
-LOAD 'pg_dbms_stats';
+\c - super_user
